@@ -1,19 +1,17 @@
 from abc import ABC
 from typing import Union, Literal
+from pydantic import BaseModel
 
-from open_mpic_core.common_domain.enum.check_type import CheckType
-from pydantic import BaseModel, Field
-
-from open_mpic_core.common_domain.check_response import CaaCheckResponse, DcvCheckResponse
-from open_mpic_core.mpic_coordinator.domain.mpic_orchestration_parameters import MpicEffectiveOrchestrationParameters
-from open_mpic_core.mpic_coordinator.domain.mpic_orchestration_parameters import MpicRequestOrchestrationParameters
-from open_mpic_core.common_domain.check_parameters import CaaCheckParameters, DcvCheckParameters
-from typing_extensions import Annotated
+from open_mpic_core import CheckType
+from open_mpic_core import CaaCheckResponse, DcvCheckResponse, CaaCheckParameters, DcvCheckParameters
+from open_mpic_core import MpicRequestOrchestrationParameters, MpicEffectiveOrchestrationParameters
 
 
 class BaseMpicResponse(BaseModel, ABC):
     request_orchestration_parameters: MpicRequestOrchestrationParameters | None = None
     actual_orchestration_parameters: MpicEffectiveOrchestrationParameters | None = None
+    check_type: CheckType
+    domain_or_ip_target: str | None = None
     is_valid: bool | None = False
     trace_identifier: str | None = None
 
